#include "gfx.h"

#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>

/* offset from top */
#define SCREEN_OFFSET 	32

/* grid sqare len */
#define SQUARE_LEN		30

/* line used to draw the grid */
#define LINE_LEN		98
#define LINE_WID		2

/* start of the grid */
#define INITIAL_X		15
#define INITIAL_Y		15

/* colors */
#define COLOR_P1		ST7735_RED
#define COLOR_P2		ST7735_GREEN
#define COLOR_SEL		ST7735_BLUE
#define COLOR_RESET		ST7735_WHITE

/* bot sqare values */
#define MIN_VALUE 	0
#define MAX_VALUE	99

/* game grid encoding */
typedef enum {
	type_unoccupied,
	type_player1,
	type_player2 = 3
} SqareType;

/* player encoding */
typedef enum {
	player1 = 1,
	player2 = 3
} MovingPlayer;

/* gamestate encoding */
typedef enum {
	game_select,
	game_play,
	game_end
} GameState;

/* game result encoding */
typedef enum {
	result_draw,
	result_wonP1,
	result_wonP2,
	result_stillPlay
} GameResult;

typedef enum {
	mode_2P,
	mode_AI
} GameMode;

/* global variables */
uint8_t selectedLine = 0, selectedCol = 0;
int gameMatix[3][3];
int currentPlayer;
GameState state;
GameResult result;
GameMode mode;

/* translate into diplay coordonates */
uint8_t getDisplayX(uint8_t x)
{
	return x;
}

/* translate into diplay coordonates */
uint8_t getDisplayY(uint8_t y)
{
	return y + SCREEN_OFFSET;
}

/* draw the TIC-TAC-TOE grid */
void drawGrid()
{
	fillRect(getDisplayX(INITIAL_X)								   , getDisplayY(INITIAL_Y)								   , LINE_LEN, LINE_WID, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X)								   , getDisplayY(INITIAL_Y +     SQUARE_LEN +     LINE_WID), LINE_LEN, LINE_WID, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X)								   , getDisplayY(INITIAL_Y + 2 * SQUARE_LEN + 2 * LINE_WID), LINE_LEN, LINE_WID, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X)								   , getDisplayY(INITIAL_Y + 3 * SQUARE_LEN + 3 * LINE_WID), LINE_LEN, LINE_WID, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X)								   , getDisplayY(INITIAL_Y)								   , LINE_WID, LINE_LEN, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X +     SQUARE_LEN +     LINE_WID), getDisplayY(INITIAL_Y)								   , LINE_WID, LINE_LEN, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X + 2 * SQUARE_LEN + 2 * LINE_WID), getDisplayY(INITIAL_Y)								   , LINE_WID, LINE_LEN, ST7735_BLACK);
	fillRect(getDisplayX(INITIAL_X + 3 * SQUARE_LEN + 3 * LINE_WID), getDisplayY(INITIAL_Y)								   , LINE_WID, LINE_LEN, ST7735_BLACK);
}

/* draw the content for a square */
void drawGameSqare(uint8_t line, uint8_t col, SqareType type)
{
	uint16_t color;
	
	/* don't draw the selected sqare */
	if (line == selectedLine && col == selectedCol)
		return;
	

	if (type == type_unoccupied)
		color = COLOR_RESET;
	else if (type == type_player1)
		color = COLOR_P1;
	else
		color = COLOR_P2;
	
	fillRect(
		getDisplayX(INITIAL_X + line * SQUARE_LEN + (line + 1) * LINE_WID), 
		getDisplayY(INITIAL_Y + col * SQUARE_LEN + (col + 1) * LINE_WID),
		SQUARE_LEN,
		SQUARE_LEN,
		color
	);
}

/* draw the select square */
void drawSelectSqare(uint8_t line, uint8_t col)
{
	fillRect(
		getDisplayX(INITIAL_X + line * SQUARE_LEN + (line + 1) * LINE_WID), 
		getDisplayY(INITIAL_Y + col * SQUARE_LEN + (col + 1) * LINE_WID),
		SQUARE_LEN,
		SQUARE_LEN,
		COLOR_SEL
	);
}

/* draw all sqares */
void drawGame()
{
	int i, j;
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			drawGameSqare(i, j, gameMatix[i][j]);
}

/* init game variables */
void initGame()
{
	int i,j;

	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			gameMatix[i][j] = type_unoccupied;

	state = game_select;
	currentPlayer = player1;
	result = result_stillPlay;
	mode = mode_AI;
}

void drawResultDraw()
{
	/* D */
	fillRect(getDisplayX(5) , getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(11), getDisplayY(30),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(11), getDisplayY(78),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(14), getDisplayY(36),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(14), getDisplayY(72),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(17), getDisplayY(42),  3, 36, ST7735_BLACK);
	fillRect(getDisplayX(20), getDisplayY(48),  3, 24, ST7735_BLACK);

	/* R */
	fillRect(getDisplayX(26), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(30),  9, 12, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(60),  9, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(36),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(54),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(44), getDisplayY(39),  3, 24, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(72),  6,  6, ST7735_BLACK);
	fillRect(getDisplayX(38), getDisplayY(78),  6,  6, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(84),  6,  6, ST7735_BLACK);

	/* A */
	fillRect(getDisplayX(50), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(74), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(50), getDisplayY(30), 24,  6, ST7735_BLACK);
	fillRect(getDisplayX(50), getDisplayY(72), 24,  6, ST7735_BLACK);

	/* W */
	fillRect(getDisplayX(86) , getDisplayY(30),  6, 48, ST7735_BLACK);
	fillRect(getDisplayX(98) , getDisplayY(30),  6, 48, ST7735_BLACK);
	fillRect(getDisplayX(110), getDisplayY(30),  6, 48, ST7735_BLACK);
	fillRect(getDisplayX(89) , getDisplayY(78), 24,  6, ST7735_BLACK);
	fillRect(getDisplayX(92) , getDisplayY(84),  8,  6, ST7735_BLACK);
	fillRect(getDisplayX(101), getDisplayY(84),  8,  6, ST7735_BLACK);

	_delay_ms(7000);
	initGame();
}

void drawResultP1()
{
	/* P */
	fillRect(getDisplayX(26), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(30),  9, 12, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(60),  9, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(36),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(54),  3, 12, ST7735_BLACK);
	fillRect(getDisplayX(44), getDisplayY(39),  3, 24, ST7735_BLACK);
	
	/* 1 */
	fillRect(getDisplayX(74), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(68), getDisplayY(42),  6, 12, ST7735_BLACK);
	fillRect(getDisplayX(62), getDisplayY(48),  3,  6, ST7735_BLACK);
	fillRect(getDisplayX(65), getDisplayY(48),  3,  6, ST7735_BLACK);
	fillRect(getDisplayX(62), getDisplayY(80), 30, 12, ST7735_BLACK);

	_delay_ms(7000);
	initGame();
}

void drawResultP2()
{
	/* P */
	fillRect(getDisplayX(26), getDisplayY(30), 6, 60, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(30), 9, 12, ST7735_BLACK);
	fillRect(getDisplayX(32), getDisplayY(60), 9, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(36), 3, 12, ST7735_BLACK);
	fillRect(getDisplayX(41), getDisplayY(54), 3, 12, ST7735_BLACK);
	fillRect(getDisplayX(44), getDisplayY(39), 3, 24, ST7735_BLACK);
	
	/* 2 */
	fillRect(getDisplayX(60), getDisplayY(30), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(60), getDisplayY(56), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(60), getDisplayY(82), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(60), getDisplayY(56),  6, 30, ST7735_BLACK);
	fillRect(getDisplayX(78), getDisplayY(30),  6, 32, ST7735_BLACK);

	_delay_ms(7000);
	initGame();
}

void drawStartScreen()
{
	/* A */
	fillRect(getDisplayX(5) , getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(29), getDisplayY(30),  6, 60, ST7735_BLACK);
	fillRect(getDisplayX(5) , getDisplayY(30), 24,  6, ST7735_BLACK);
	fillRect(getDisplayX(5) , getDisplayY(72), 24,  6, ST7735_BLACK);

	/* I */
	fillRect(getDisplayX(38), getDisplayY(30),  6, 60, ST7735_BLACK);
	
	/* VLINE */
	fillRect(getDisplayX(61),  getDisplayY(0),  6, 128, ST7735_BLACK);
	
	/* 2 */
	fillRect(getDisplayX(70), getDisplayY(30), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(70), getDisplayY(56), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(70), getDisplayY(82), 21,  6, ST7735_BLACK);
	fillRect(getDisplayX(70), getDisplayY(56),  6, 30, ST7735_BLACK);
	fillRect(getDisplayX(88), getDisplayY(30),  6, 32, ST7735_BLACK);
	
	/* P */
	fillRect(getDisplayX(100), getDisplayY(30), 6, 60, ST7735_BLACK);
	fillRect(getDisplayX(106), getDisplayY(30), 9, 12, ST7735_BLACK);
	fillRect(getDisplayX(106), getDisplayY(60), 9, 12, ST7735_BLACK);
	fillRect(getDisplayX(115), getDisplayY(36), 3, 12, ST7735_BLACK);
	fillRect(getDisplayX(115), getDisplayY(54), 3, 12, ST7735_BLACK);
	fillRect(getDisplayX(118), getDisplayY(39), 3, 24, ST7735_BLACK);

	_delay_ms(3000);
}

/* get game result */
GameResult checkResult()
{
	int allOccupied = 1, i, j;

	/* check all columns */
	for (i = 0; i < 3; i++) {
		if (gameMatix[i][0] == type_player1 && gameMatix[i][1] == type_player1 && gameMatix[i][2] == type_player1)
			return result_wonP1;
		if (gameMatix[i][0] == type_player2 && gameMatix[i][1] == type_player2 && gameMatix[i][2] == type_player2)
			return result_wonP2;
	}

	/* check all lines */
	for (j = 0; j < 3; j++) {
		if (gameMatix[0][j] == type_player1 && gameMatix[1][j] == type_player1 && gameMatix[2][j] == type_player1)
			return result_wonP1;
		if (gameMatix[0][j] == type_player2 && gameMatix[1][j] == type_player2 && gameMatix[2][j] == type_player2)
			return result_wonP2;
	}

	/* check first diagonal */
	if (gameMatix[0][0] == type_player1 && gameMatix[1][1] == type_player1 && gameMatix[2][2] == type_player1)
		return result_wonP1;
	if (gameMatix[0][0] == type_player2 && gameMatix[1][1] == type_player2 && gameMatix[2][2] == type_player2)
		return result_wonP2;

	/* check second diagonal */
	if (gameMatix[0][2] == type_player1 && gameMatix[1][1] == type_player1 && gameMatix[2][0] == type_player1)
		return result_wonP1;
	if (gameMatix[0][2] == type_player2 && gameMatix[1][1] == type_player2 && gameMatix[2][0] == type_player2)
		return result_wonP2;

	/* check if all sqares are filled*/
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++)
			if (gameMatix[i][j] == type_unoccupied)
				allOccupied = 0;

	if (allOccupied == 0)
		return result_stillPlay;

	return result_draw;
}

/* Give a value for a sqare */
int getSqareValue(int line, int col)
{
	int value = 0;

	/* check if ai can complete the current line */
	if (gameMatix[line][0] != type_player1 && gameMatix[line][1] != type_player1 && gameMatix[line][2] != type_player1)
		value++;

	/* check if ai can complete the current column */
	if (gameMatix[0][col] != type_player1 && gameMatix[1][col] != type_player1 && gameMatix[2][col] != type_player1)
		value++;

	/* check if sqare is on the first diagonal */
	if (line == col)
		if (gameMatix[0][0] != type_player1 && gameMatix[1][1] != type_player1 && gameMatix[2][2] != type_player1)
			value++;

	/* check if sqare is on the second diagonal */
	if (line + col == 2)
		if (gameMatix[0][2] != type_player1 && gameMatix[1][1] != type_player1 && gameMatix[2][0] != type_player1)
			value++;

	return value;
}

/* return max_value if lost is prevented or min_value otherwise */
int preventLoss(int line, int col)
{
	int cellScore = 0;

	/* check line */
	cellScore = gameMatix[line][0] + gameMatix[line][1] + gameMatix[line][2];
	if (cellScore == 2 * type_player1)
		return MAX_VALUE;

	/* check column */
	cellScore = gameMatix[0][col] + gameMatix[1][col] + gameMatix[2][col];
	if (cellScore == 2 * type_player1)
		return MAX_VALUE;

	/* check if sqare is on the first diagonal */
	if (line == col) {
		cellScore = gameMatix[0][0] + gameMatix[1][1] + gameMatix[2][2];
		if (cellScore == 2 * type_player1)
			return MAX_VALUE;
	}

	/* check if sqare is on the second diagonal */
	if (line + col == 2) {
		cellScore = gameMatix[2][0] + gameMatix[1][1] + gameMatix[0][2];
		if (cellScore == 2 * type_player1)
			return MAX_VALUE;
	}

	return MIN_VALUE; 
}

/* return max_value if ai win or min_value otherwise */
int winGame(int line, int col)
{
	int cellScore = 0;

	/* check line */
	cellScore = gameMatix[line][0] + gameMatix[line][1] + gameMatix[line][2];
	if (cellScore == 2 * type_player2)
		return MAX_VALUE;

	/* check column */
	cellScore = gameMatix[0][col] + gameMatix[1][col] + gameMatix[2][col];
	if (cellScore == 2 * type_player2)
		return MAX_VALUE;

	/* check if sqare is on the first diagonal */
	if (line == col) {
		cellScore = gameMatix[0][0] + gameMatix[1][1] + gameMatix[2][2];
		if (cellScore == 2 * type_player2)
			return MAX_VALUE;
	}

	/* check if sqare is on the second diagonal */
	if (line + col == 2) {
		cellScore = gameMatix[2][0] + gameMatix[1][1] + gameMatix[0][2];
		if (cellScore == 2 * type_player2)
			return MAX_VALUE;
	}

	return MIN_VALUE; 
}

void aiMakeMove()
{
	int bestLine = 0, bestCol = 0, i,j, bestScore = -1, score;

	/* compute a score for each posible sqare */
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			if (gameMatix[i][j] != type_unoccupied)
				continue;

			/* try to win */
			score = winGame(i, j);
			if (score > bestScore) {
				bestScore = score;
				bestLine = i;
				bestCol = j;
			}

			/* try not to lose */
			score = preventLoss(i, j);
			if (score > bestScore) {
				bestScore = score;
				bestLine = i;
				bestCol = j;
			}

			/* whatever */
			score = getSqareValue(i, j);
			if (score > bestScore) {
				bestScore = score;
				bestLine = i;
				bestCol = j;
			}
		}
	}

	/* make the move on the best sqare */
	gameMatix[bestLine][bestCol] = type_player2;

	/* check game result */
	result = checkResult();
	if (result != result_stillPlay)
		state = game_end;
}

/* Button from portA (P1 buttons) */
ISR(PCINT0_vect)
{
	if (state == game_select) {
		state = game_play;
		mode = mode_AI;
		fillScreen(COLOR_RESET);
		_delay_ms(500);
		return;
	}

	// if (state == game_end) {
	// 	initGame();
	// 	return;
	// }

	// P1 LEFT
	if((PINA & (1 << PA0)) == 0 && currentPlayer == player1)
	{
		if (selectedLine == 0)
			selectedLine = 2;
		else
			selectedLine--;
			
		_delay_ms(10); /* debounce */
	}
	
	// P1 UP
	if((PINA & (1 << PA1)) == 0 && currentPlayer == player1)
	{
		if (selectedCol == 0)
			selectedCol = 2;
		else
			selectedCol--;
			
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P1 DOWN
	if((PINA & (1 << PA2)) == 0 && currentPlayer == player1)
	{
		selectedCol = (selectedCol + 1) % 3;
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P1 RIGHT
	if((PINA & (1 << PA3)) == 0 && currentPlayer == player1)
	{
		selectedLine = (selectedLine + 1) % 3;
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P1 SELECT
	if((PINA & (1 << PA4)) == 0 && currentPlayer == player1)
	{
		if (gameMatix[selectedLine][selectedCol] == type_unoccupied) {
			gameMatix[selectedLine][selectedCol] = currentPlayer;
			result = checkResult();
			if (result != result_stillPlay) {
				state = game_end;
			}
			else {
				if (mode == mode_2P)
					currentPlayer = player2;
				else
					aiMakeMove();
			}
		}

		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P2 SELECT
	if((PINA & (1 << PA5)) == 0 && currentPlayer == player2)
	{
		if (gameMatix[selectedLine][selectedCol] == type_unoccupied) {
			gameMatix[selectedLine][selectedCol] = currentPlayer;
			result = checkResult();
			if (result != result_stillPlay)
				state = game_end;
			else
				currentPlayer = player1;
		}

		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}
}

/* Button from portD (P2 buttons) */
ISR(PCINT1_vect)
{

	if (state == game_select) {
		state = game_play;
		mode = mode_2P;
		fillScreen(COLOR_RESET);
		_delay_ms(500);
		return;
	}

	// P2 LEFT
	if((PINB & (1 << PB0)) == 0 && currentPlayer == player2)
	{
		if (selectedLine == 0)
			selectedLine = 2;
		else
			selectedLine--;
			
		_delay_ms(10); /* debounce */
	}
	
	// P2 UP
	if((PINB & (1 << PB1)) == 0 && currentPlayer == player2)
	{
		if (selectedCol == 0)
			selectedCol = 2;
		else
			selectedCol--;
			
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P2 DOWN
	if((PINB & (1 << PB2)) == 0 && currentPlayer == player2)
	{
		selectedCol = (selectedCol + 1) % 3;
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}

	// P2 RIGHT
	if((PINB & (1 << PB3)) == 0 && currentPlayer == player2)
	{
		selectedLine = (selectedLine + 1) % 3;
		PORTD ^= (1<<PD7);
		_delay_ms(10); /* debounce */
	}
}

void initInput()
{
	/* set ports as input and activate the pull-down rezistor portA*/
	DDRA &= ~((1 << PA0) | (1 << PA1) | (1<<PA2) | (1<<PA3) | (1<<PA4) | (1<<PA5));
	PORTA |= ((1 << PA0) | (1 << PA1) | (1<<PA2) | (1<<PA3) | (1<<PA4) | (1<<PA5));

	/* activate interups for portA(0..5) */
	PCICR  |= (1 << PCIE0);
	PCMSK0 |= (31 << PCINT0) | (1 << PCINT5);

	/* set ports as input and activate the pull-down rezistor portB*/
	DDRB &= ~((1 << PB0) | (1 << PB1) | (1<<PB2) | (1<<PB3));
	PORTB |= ((1 << PB0) | (1 << PB1) | (1<<PB2) | (1<<PB3));

	/* activate interups for portB(0..3) */
	PCICR  |= (1 << PCIE1);
	PCMSK1 |= (15 << PCINT8);
}

int main(void){
	// init the 1.8 lcd display
	init();
	initInput();
	initGame();
	sei();

	constructor(_width,_height);
	
	DDRD |= (1<<PD7);
	PORTD |= (1<<PD7);

	while(1){
		setRotation(0);
		if (state == game_select) {
			fillScreen(COLOR_RESET);
			_delay_ms(500);
			drawStartScreen();
		} else if (state == game_play) {
			drawGrid();
			drawGame();
			drawSelectSqare(selectedLine, selectedCol);
		} else if (state == game_end) {
			fillScreen(COLOR_RESET);
			if (result == result_draw)
				drawResultDraw();
			if (result == result_wonP1)
				drawResultP1();
			if (result == result_wonP2)
				drawResultP2();
		}
	}

	return 0;
}
